/***** rank: counterexample.c *****/

/* Written by Seth Fogarty, Rice University, USA                          */
/* Based on code written by Laurent Doyen, ULB, Belgium                   */
/* Portions copyright (c) 2006  Laurent Doyen                             */
/* Portions copyright (c) 2007  Seth Fogarty                              */
/*                                                                        */
/* This program is free software; you can redistribute it and/or modify   */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation; either version 2 of the License, or      */
/* (at your option) any later version.                                    */
/* This program is distributed in the hope that it will be useful,        */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */
/*                                                                        */
/* You should have received a copy of the GNU General Public License      */
/* along with this program; if not, write to the Free Software            */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA*/

//If defined, the set is not necesarilly a set of antichains. Not verified that
//this does not change the logic of comparing them. Not changed is the
//generation of antichains in the 'Pre' operation, nor the comparing of sets
//using the subsumption relation instead of equality.
//#define NOSUBSUMPTION

#include "rank.h"

typedef struct _Record{
  Pair* pair;
  struct _Record* succ;
  Letter succ_letter;
  bool visited;
  bool begins_loop;
} Record;

// List with first element dummy
typedef struct _RAntichain{
  Record* record;
  struct _RAntichain* next;
} RAntichain;


// SF: Constructors and destructors

Record* createRecord(Set s, Set o, Record* succ, int letter)
{
  Record* result = (Record*) malloc(sizeof(Record));
  result->pair = createPair(s,o);
  result->succ = succ;
  result->succ_letter = letter;
  result->visited = false;
  result->begins_loop= false;
  return result;
}

Record* createRecordFromPair(Pair* pair, Record* succ, int letter)
{
  Record* result = (Record*) malloc(sizeof(Record));
  result->pair = pair;
  result->succ = succ;
  result->succ_letter = letter;
  result->visited = false;
  result->begins_loop= false;
  return result;
}

Record* createRPair(Set s, Set o, Record* succ, int letter)
{
  Record* result = (Record*) malloc(sizeof(Record));
  result->pair = createPair(s,o);
  result->succ = succ;
  result->succ_letter = letter;
  result->visited=false;
  result->begins_loop= false;
  return result;
}


Record* copyRecord(Record* t)
{
  Record* result = (Record*) malloc(sizeof(Record));
  result->pair = copyPair(t->pair);
  result->succ = t->succ; 
  result->succ_letter = t->succ_letter;
  result->visited=false;
  result->begins_loop= false;
  return result;
}


RAntichain* createRAntichain()
{
  RAntichain* result = (RAntichain*) malloc(sizeof(RAntichain));
  result->record=NULL;
  result->next=NULL;
  return result;
}

void deleteRecord(Record* p)
{
  if (p != NULL)
  {
    deletePair(p->pair);
    free(p);
    p = NULL;
  }

}

void deleteRAntichain(RAntichain* Q)
{
  RAntichain* Old;
  while(Q != NULL)
  {
    deleteRecord(Q->record);
    Old = Q; 
    Q=Q->next;
    free(Old);
    Old = NULL;
  }
}

/*SF: This goes through and deletes all the antichain structs that were holding
 * the pairs together, but leaves the pairs intact. If you do not use R after
 * appendRAntichain, call this. */
void deleteRAntichainSpine(RAntichain* Q)
{
  RAntichain* Old;
  while(Q != NULL)
  {
    Old = Q; 
    Q=Q->next;
    free(Old);
    Old=NULL;
  }
}

//y=[[s,s],[s,owe]]  # y=top
/* SF: The top antichain is no longer small: Specifically, pairs with
 * different a's are incomprable. Thus we now have a 2*sizeA top antichain. */

/* we shouldn't ever create this.
RAntichain* topRAntichain()
{
  State i;
  RAntichain* y = createRAntichain();
  RAntichain* cur = y;
  if(containment)
  {
    for(i = 0; i < sizeA; i++)
    {
      cur->next=createRAntichain();
      cur->next->record = createRecord(i, maximalSet(), maximalSet(), NULL, -1);
      cur->next->next = createRAntichain();
      cur->next->next->pair = createRecord(i, maximalSet(), emptySet(), NULL, -1);
      cur = cur->next->next;
    }
  }
  else
  {
    y->next = createRAntichain();
    y->next->pair = createPair(maximalSet(), maximalSet());
    y->next->next = createRAntichain();
    y->next->next->pair = createPair(maximalSet(), emptySet());
  }
  return y;
}
*/


/*SF: Similarly, our initial antichain is now duplicated for every initial
 * state in A. It would be nice to know if the sets created really need to be
 * recreated each time: if they are write only (which I suspect they are), then
 * we can greatly simplify this and the above via memory-sharing.*/

RAntichain* initialRAntichain()
{
  RAntichain* y = createRAntichain();
  State i;
  RAntichain* cur = y;
  if(containment)
  {
     for(i = 0; i < sizeA; i++)
     {
       if(initialA[i])
       {
         cur->next = createRAntichain();
         cur->next->record= createRecord(initialSet(), emptySet(), NULL, -1);
         cur = cur->next;
       }
     }
  }
  else
  {
     y->next = createRAntichain();
     y->next->record= createRPair(initialSet(), emptySet(), NULL, -1);
  }
  return y;
}


Record* intersectRecord(Record *t1, Record *t2)
{
  Record* result = NULL;
  createRecordFromPair(intersectPair(t1->pair,t2->pair), NULL, -1);
  fprintf(stderr, "Intersection of records performed and not well defined.");
  return result;
}

int rgetMaxRank(RAntichain* Q)
{
  int v;
  int result=0;
  for(Q=Q->next; Q != NULL; Q=Q->next)
  {
    for(v=0; v<sizeB; v++)
    {
      int rank = Q->record->pair->o[v];
      if (rank <= maxRank)
      {
        if (rank > result) result=rank;
      }
    }
    for(v=0; v<sizeB; v++)
    {
      int rank = Q->record->pair->s[v];
      if (rank <= maxRank)
      {
        if (rank > result) result=rank;
      }
    }

  }
  return result;
}

void rupdateLargestRank(RAntichain* Q)
{
  int rank=rgetMaxRank(Q);
  if (rank > largestRank) largestRank = rank;
}


int sizeRAntichain(RAntichain* Q)
{
  int result=0;
  for(Q=Q->next; Q != NULL; Q=Q->next)
  {
    result++;
  }
  return result;
}

bool isEmptyRAntichain(RAntichain* Q)
{
  return(Q->next == NULL);
}


int insertRAntichain(RAntichain* Q, Record* record)
{
  num_inserts++;
  assert(Q->record == NULL);
  RAntichain* previous =Q;
  //printf("* * * Entering insertRAntichain * * *}

  for(Q=Q->next; Q != NULL;)
  {
    // printPair(Q->pair);
   // if (LEQpair(record->pair, Q->record->pair))
    if (EQpair(record->pair, Q->record->pair))
    {
      if(Q->record->succ==NULL) {
         Q->record->succ = record->succ;
         Q->record->succ_letter = record->succ_letter;
      }
      deleteRecord(record);
      num_subsumed++;
      return -1;
    }
    else
    {
      Q=Q->next;
      previous = previous->next;
    }
  }
  previous->next = createRAntichain(); //SF: Q is null  here, so safe. Note we did not insert above,
  previous->next->record = record;
  return 0;
}

RAntichain* intersectRAntichain(RAntichain* Q, RAntichain* P)
{
  RAntichain* Pstart=P;
  RAntichain* result = createRAntichain();
  Record* intersection;
  for(Q=Q->next; Q != NULL; Q=Q->next)
  {
    P=Pstart;
    for(P=P->next; P != NULL; P=P->next)
    {
      assert(P->record != NULL);
      assert(Q->record != NULL);
      intersection=intersectRecord(P->record, Q->record);
      if (intersection != NULL) insertRAntichain(result, intersection);
    }
  }
  return result;
}

/* SF: fix by pushing R = R -> next down into each branch, and using a
 * temporary variable to hold pre-increment R here, and then delete as last
 * step. 
*/

/*SF: I believe this is DESTRUCTIVE on R, but is not totally so. 
 *In the CPRE's, that is certainly what we want to do. It appears that in the
 * least fixed point, however, we want R to be intact but modified. */

 /*This copies pointers to the pairs of R, thus after calling this function call only deleteRAntichainSpine(R) */
 
void appendRAntichain(RAntichain* Q, RAntichain*  R)
{
  RAntichain* previous= R;
  RAntichain* Old;
  for(R=R->next; R!=NULL;)
  {
    if (insertRAntichain(Q, (R->record)) == -1)
    {
      //printf("#");
      previous->next = R->next;  
      Old = R;
      R=R->next;
      free(Old); 
      /* we do NOT want to delete the antichain here, we have already freed
       * R->pair, we don't want to free R-> next, so we simply free R.*/
    }
    else
    {
      R=R->next; 
      previous = previous->next;
    }
  }
}



// SF: Begin of heavy hitting logic functions. Here is where I miss OCaml. 

/* SF: The boolvalue acceptA tells us if we should restrict ourselves to
 * pairs containing an accepting state in A. it is the analogue to
 * PreRecordGoalA */
RAntichain* preRecord(Record* p)
{
  Set preo;
  Set pres;
  Set o=p->pair->o;
  Set s=p->pair->s;
  RAntichain* result;
  int j, v, w, rank, temp;
  Letter sigma;
  
  result = createRAntichain();
  preo = emptySet();
  pres = emptySet();
  
  for(sigma=0; sigma<num_letters; sigma++)
  {
    clearSet(preo); 
    for(v=0; v<sizeB; v++)
    {
      // printf("v = %d", v);
      rank = 0;
      for(j=0; postB[sigma][v][j] != -1; j++)
      {
        w=postB[sigma][v][j];
        if (acceptingB[w])
        {
          if (o[w] > rank)
          {
            rank = o[w];
          }
        }
        else
        {
          temp = min(ceilOdd(s[w]), o[w]);
          if (temp > rank)
          {
            rank = temp;
          }
        }
      }
      if (acceptingB[v])
        if (rank%2 == 1) rank++;
      preo[v] = rank;
    }
  
    /* the above generates pre-o, this generates pre-s if pre-o is non-empty. */
    if (!isEmptySet(preo))
    {
      clearSet(pres);
      for(v=0; v<sizeB; v++)
      {
        rank = 0;
        for(j=0; postB[sigma][v][j] != -1; j++)
        {
          w=postB[sigma][v][j];
          if (s[w] > rank)
            rank = s[w];
        }
        if (acceptingB[v])
          if (rank%2 == 1) rank++;
        pres[v] = rank;
      }
      /* The predecessors of a state a,s,o is any state(a',s',o') such that a'->a and blah blah */
      //for(j=0; preA[sigma][a][j] != -1; j++)
      //{
       // temp = preA[sigma][a][j]; 
        insertRAntichain(result, createRecord(copySet(pres), copySet(preo), p, sigma));
      //}
    }


    //for(j=0; preA[sigma][a][j] != -1; j++)
    //{
     //temp = preA[sigma][a][j]; 
     insertRAntichain(result, createRecord(copySet(preo), emptySet(), p, sigma)); 
    //}
  }
  free(pres);
  free(preo); 
  return result;
}

RAntichain* prePairRecord(Record* p)
{
  Set preo;
  Set o;
  Set pres;
  Set s;
  RAntichain* result;
  int j;
  int v,w;
  int rank;
  int temp;
  Letter sigma;
  
  result = createRAntichain();
  o=p->pair->o;
  s=p->pair->s;
  
  for(sigma=0; sigma<num_letters; sigma++)
  {
    preo = emptySet();
    for(v=0; v<sizeB; v++)
    {
      // printf("v = %d", v);
      rank = 0;
      for(j=0; postB[sigma][v][j] != -1; j++)
      {
        w=postB[sigma][v][j];
        if (acceptingB[w])
        {
          if (o[w] > rank)
          {
            rank = o[w];
          }
        }
        else
        {
          temp = min(ceilOdd(s[w]), o[w]);
          if (temp > rank)
          {
            rank = temp;
          }
        }
      }
      if (acceptingB[v])
        if (rank%2 == 1) rank++;
      preo[v] = rank;
    }
  
    if (!isEmptySet(preo))
    {
      pres = emptySet();
      for(v=0; v<sizeB; v++)
      {
        rank = 0;
        for(j=0; postB[sigma][v][j] != -1; j++)
        {
          w=postB[sigma][v][j];
          if (s[w] > rank)
            rank = s[w];
        }
        if (acceptingB[v])
          if (rank%2 == 1) rank++;
        pres[v] = rank;
      }
      
      insertRAntichain(result, createRPair(pres, copySet(preo), p, sigma));
    }
    insertRAntichain(result, createRPair(preo, emptySet(), p, sigma)); 
  }

  return result;
}

RAntichain* rCPre(RAntichain* Q)
{
  RAntichain* result = createRAntichain();
  RAntichain* cpreQ;
  for(Q=Q->next; Q != NULL; Q=Q->next)
  {
    // if (DEBUG) {printf("CPre of "); printPair(Q->pair);}
    if(containment) 
        cpreQ = preRecord(Q->record);
    else 
        cpreQ = prePairRecord(Q->record);
    appendRAntichain(result, cpreQ);
    //SF: Need to free more than just cpreQ.
    // free(cpreQ);
    deleteRAntichainSpine(cpreQ);
  }
  return result;
}



/*We only compute pre with a Goal in the first step, and we take in a pair,
 * not a record. */
RAntichain* preRecordGoalB(Pair* p)
{
  Set preo;
  Set o;
  Set s;
  RAntichain* result;
  int j;
  int v,w;
  int rank;
  int temp;
  Letter sigma;
  
  result = createRAntichain();
  o=p->o;
  s=p->s;
  preo = emptySet();
  
  for(sigma=0; sigma<num_letters; sigma++)
  {
    clearSet(preo);
    for(v=0; v<sizeB; v++)
    {
      // printf("v = %d", v);
      rank = 0;
      for(j=0; postB[sigma][v][j] != -1; j++)
      {
        w=postB[sigma][v][j];
        if (acceptingB[w])
        {
          if (o[w] > rank)
          {
            rank = o[w];
          }
        }
        else
        {
          temp = min(ceilOdd(s[w]), o[w]);
          if (temp > rank)
          {
            rank = temp;
          }
        }
      }
      if (acceptingB[v])
        if (rank%2 == 1) rank++;
      preo[v] = rank;
    }
    // if (checkRelation(o,s))
      // printf("This test was useful");
      // pause();
   if (containment)
   {
      //for(j=0; preA[sigma][a][j] != -1; j++)
      //{
       // insertRAntichain(result, createRecord(preA[sigma][a][j], copySet(preo), emptySet(), NULL, -1)); 
      //}
      printf("Counterexamples not supported for containment.\n");
   }
   else
   {
    insertRAntichain(result, createRPair(copySet(preo), emptySet(), NULL, -1)); 
   }
  }
  free(preo);
  return result;
}

//RAntichain* preRecordGoalA(Pair* p)
//{
//  Set preo;
//  Set pres;
//  Set o=p->o;
//  Set s=p->s;
//  RAntichain* result;
//  int j, v, w, rank, temp;
//  Letter sigma;
//  
//  result = createRAntichain();
//  preo = emptySet();
//  pres = emptySet();
//  
//  for(sigma=0; sigma<num_letters; sigma++)
//  {
//    clearSet(preo); 
//    for(v=0; v<sizeB; v++)
//    {
//      // printf("v = %d", v);
//      rank = 0;
//      for(j=0; postB[sigma][v][j] != -1; j++)
//      {
//        w=postB[sigma][v][j];
//        if (acceptingB[w])
//        {
//          if (o[w] > rank)
//          {
//            rank = o[w];
//          }
//        }
//        else
//        {
//          temp = min(ceilOdd(s[w]), o[w]);
//          if (temp > rank)
//          {
//            rank = temp;
//          }
//        }
//      }
//      //if(v==12) printf("found v=12: rank=%d, acceptingB[v]=%d\n", rank, acceptingB[v]);
//      if (acceptingB[v])
//        if (rank%2 == 1) rank++;
//      //if(v==12) printf("found v=12: rank now=%d\n", rank);
//      preo[v] = rank;
//    }
//    //if(report && (sigma==1)) {printSet(preo); printf("\n");}
//  
//    /* the above generates pre-o, this generates pre-s if pre-o is non-empty. */
//    if (!isEmptySet(preo))
//    {
//      clearSet(pres);
//      for(v=0; v<sizeB; v++)
//      {
//        rank = 0;
//        for(j=0; postB[sigma][v][j] != -1; j++)
//        {
//          w=postB[sigma][v][j];
//          if (s[w] > rank)
//            rank = s[w];
//        }
//        if (acceptingB[v])
//          if (rank%2 == 1) rank++;
//        pres[v] = rank;
//      }
//      /* The predecessors of a state a,s,o is any state(a',s',o') such that a'->a and blah blah */
//      //for(j=0; preA[sigma][a][j] != -1; j++)
//      //{
//       // temp = preA[sigma][a][j]; 
//        //if(acceptingA[temp])
//        //{
//          insertRAntichain(result, createRecord(copySet(pres), copySet(preo), NULL, -1));
//        //}
//      //}
//    }
//
//
//    for(j=0; preA[sigma][a][j] != -1; j++)
//    {
//     temp = preA[sigma][a][j]; 
//     if(acceptingA[temp])
//     {
//       insertRAntichain(result, createRecord(temp, copySet(preo), emptySet(), NULL, -1)); 
//     }
//   }
//  }
//  free(pres);
//  free(preo); 
//  return result;
//}


//This does not affect Q in any way, use normal deleteRAntichain
//RAntichain* rCPreGoalA(Antichain* Q)
//{
//  RAntichain* result = createRAntichain();
//  RAntichain* cpreQ;
//  for(Q=Q->next; Q != NULL; Q=Q->next)
//  {
//    // if (DEBUG) {printf("CPre of "); printPair(Q->pair);}
//    cpreQ = preRecordGoalA(Q->pair);
//    appendRAntichain(result, cpreQ);
//    //SF: Need to free more than just cpreQ
//    //free(cpreQ);
//    deleteRAntichainSpine(cpreQ);
//  }
//  return result;
//}

//This does not affect Q in any way, use normal deleteRAntichain
RAntichain* rCPreGoalB(Antichain* Q)
{
  RAntichain* result = createRAntichain();
  RAntichain* cpreQ;
  for(Q=Q->next; Q != NULL; Q=Q->next)
  {
    // if (DEBUG) {printf("CPre of "); printPair(Q->pair);}
    cpreQ = preRecordGoalB(Q->pair);
    appendRAntichain(result, cpreQ);
    //SF: Need to free more than just cpreQ
    //free(cpreQ);
    deleteRAntichainSpine(cpreQ);
  }
  return result;
}

bool LEQRantichain(RAntichain* Q, RAntichain* R)
{
  bool ok;
  RAntichain* saveR=R;
  for(Q=Q->next; Q != NULL; Q=Q->next)
  {
    ok=false;
    for(R=saveR->next; R != NULL; R=R->next)
    {
      if (LEQpair(Q->record->pair, R->record->pair))
      {
        ok = true;
        break;
      }
    }
    if (!ok)
    {
      return false;
    }
  }
  return true;
}

Record* IntersectionRecord(Antichain* init, RAntichain* fixpoint)
{
  RAntichain* savefixpoint=fixpoint;

  //we want to ensure that nothing in init is covered by fixpoint.
  //Thus we look for a counterexample.
  
  for(init=init->next; init != NULL; init=init->next)
  {
    for(fixpoint=savefixpoint->next; fixpoint != NULL; fixpoint=fixpoint->next)
    {
      if (LEQpair(init->pair, fixpoint->record->pair))
      {
         return fixpoint->record;
      }
    }
  }
  printf("No initial state in fixpoint!\n");
  return NULL;
}

// returns LEQantichain({pair}, Q)
bool LEQpairRAntichain(Record* record, RAntichain* Q)
{
  for(Q=Q->next; Q != NULL; Q=Q->next)
  {
    if (LEQpair(record->pair, Q->record->pair))
    {
      return true;
    }
  }
  return false;
}


//This is TOTALLY destructive of the goal. After you call this function, goal
//is completely freed.
RAntichain* rComputeLeastFixedPoint(RAntichain* goal)
{
  RAntichain* x, *old;
  RAntichain* frontier;
  //RAntichain* preRAntichain;

  frontier = goal;
  x=createRAntichain();
  appendRAntichain(x, frontier);

  while(!isEmptyRAntichain(frontier))
  {
    if (TIME_LIMIT > 0)
    {
      if (get_elapsed_time() > TIME_LIMIT) return x;
    }
    old = frontier;
    frontier = rCPre(frontier);  // memory leak ?? 
    //if (reallyVerbose)printSmartRAntichain(frontier);
    /*SF: Do we need to delete the old fronteir? I believe so, but how do we
     * know which elements to delete? I think it is safe to delete everything...*/
    deleteRAntichainSpine(old);

    appendRAntichain(x, frontier);  // this modifies frontier 
    /*SF: Specifically, it removes things in frontier that already in x. so
     * everything that is in fronteir now will NOT be in it next pass. Nice. So
     * our CPre will only get 'new' things. But when we insert something new,
     * it stays on the frontier*/
  }
  deleteRAntichainSpine(frontier);
  return x;
}

void printSmartRAntichain(RAntichain* Q)
{
  int n=0;;
  for(Q=Q->next; Q != NULL; Q=Q->next)
  {
    if(verbose) printf("%d:  ", n++);
    printSmartPair(Q->record->pair);
    /*
    if (!isEmptySet(Q->pair->o)) printf("  ");
    //for(i=0; i< 6*sizeB - 3*(sizeSet(Q->pair->s)+sizeSet(Q->pair->o)); i++) printf(" ");
    printf("{");
    for(v=0; v<sizeB; v++)
    {
      rank=0;
      if (Q->pair->o[v] <= maxRank)
      {
        rank = Q->pair->o[v];
        if (rank < Q->pair->s[v])
        {
          printf("Error in printSmartAntichain !\n Rank in o for %d: %d   in s: %d\n%d\n", v, rank, Q->pair->s[v], maxRank);
          exit(1);
        }
      }
      else if (Q->pair->s[v] <= maxRank)          
        rank = Q->pair->s[v];
      
      if (v!=0) printf(", ");
      printf("%d: %d", v, rank);
    }
    printf("}");  
    printf("\n");
    */
  }
  printf("\n---\n");
}



Word* Word_Reverse(Word* word)
{
  Word *pr = NULL;
  Word *tmp;
  while (word != NULL)
  {
    tmp = word->next;
    word->next = pr;
    pr = word;
    word = tmp;
  }
  return pr; 
}

Counterexample* generateCounterexample(Antichain* init, Antichain* y)
{

  RAntichain* yPrime, *cpreyA, *cpreyB, *fixedA, *fixedB;
  cpreyA=NULL;
  Record* current, *start;
  Counterexample* retval = (Counterexample*) malloc(sizeof(Counterexample));
  Word* temp;
  retval->prefix=NULL; 
  retval->loop=NULL; 

  if(containment)
  {
     printf("Containment currently not supported.\n");
     return retval;
     //cpreyA = rCPreGoalA(y); //SF: Pre(y) intersect accepting of A
     //and here we have our problem with containment. still ahve to figure this
     //out.
  }
  cpreyB = rCPreGoalB(y); //SF: Pre(y) intersect accepting of B
  if(containment)
  {
     printf("Containment currently not supported.\n");
     fixedA = rComputeLeastFixedPoint(cpreyA);
     fixedB = rComputeLeastFixedPoint(cpreyB);
     yPrime=intersectRAntichain(fixedA, fixedB);  //Do I now need to free fixedA and fixedB?
     deleteRAntichain(fixedA);
     deleteRAntichain(fixedB);
  }
  else
  {
     yPrime=rComputeLeastFixedPoint(cpreyB); //SF: mu x . Pre(x) union cprey
  }
  
  //free(y); Unlike 'checkContainment', we do not generate y, we are passed it, and
  //we should not get rid of it.
  start = IntersectionRecord(init, yPrime);
  current = start;
  assert(current != NULL);
  while(current->visited == false)
  {
    assert(current->succ != NULL);
    current->visited=true;
    current=current->succ;
  }
  current->begins_loop=true;
  current = start;
  while(current->begins_loop==false)
  {
    assert(current->visited);
    temp = (Word*)malloc(sizeof(Word));
    temp->letter=current->succ_letter;
    temp->next=retval->prefix;
    retval->prefix=temp;
    current=current->succ;
  }
  temp = (Word*)malloc(sizeof(Word));
  temp->letter=current->succ_letter;
  temp->next=NULL;
  retval->loop=temp;
  current=current->succ;
  while(current->begins_loop==false)
  {
    assert(current->visited);
    temp = (Word*)malloc(sizeof(Word));
    temp->letter=current->succ_letter;
    temp->next=retval->loop;
    retval->loop=temp;
    current=current->succ;
  }
  retval->prefix=Word_Reverse(retval->prefix);
  retval->loop=Word_Reverse(retval->loop);
  return retval;
}
