/***** rank: rank.h *****/

/* Written by Seth Fogarty, Rice University, USA                          */
/* Based on code written by Laurent Doyen, ULB, Belgium                   */
/* Portions copyright (c) 2006  Laurent Doyen                             */
/* Portions copyright (c) 2011  Seth Fogarty                              */
/*                                                                        */
/* This program is free software; you can redistribute it and/or modify   */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation; either version 2 of the License, or      */
/* (at your option) any later version.                                    */
/* This program is distributed in the hope that it will be useful,        */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */
/*                                                                        */
/* You should have received a copy of the GNU General Public License      */
/* along with this program; if not, write to the Free Software            */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA*/

#ifndef RANKH
#define RANKH

#define Free free
#define Malloc malloc

#include "string.h"

#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <memory.h>

#include "standard_io.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>
#include <stdbool.h>

//#define OUTPUT

// if TIME_LIMIT is negative ----> no limit of time
#define TIME_LIMIT -1.0
//#define OLDPRINT

#define false 0
#define true 1

#define SIZE 50 

//If defined, the set is not necesarilly a set of antichains. Not verified that
//this does not change the logic of comparing them. Not changed is the
//generation of antichains in the 'Pre' operation, nor the comparing of sets
//using the subsumption relation instead of equality.
//#define NOSUBSUMPTION

extern bool verbose;
extern bool reallyVerbose;

// Set = array of <n> int 
// the info in s[i] is the rank of the element <i>
typedef int *Set;
// List = array of some int, the last element is -1
// List are simple sets of elements list[i] (except the last)
typedef int *List;
typedef int Letter;
typedef int State;

typedef struct _Word{
  Letter letter;
  struct _Word* next;
} Word;

typedef struct{
  Word* prefix;
  Word* loop;
} Counterexample;


typedef struct{
  //State a;
  Set s;
  Set o;
} Pair;

// List with first element dummy
typedef struct _Antichain{
  Pair* pair;
  struct _Antichain* next;
} Antichain;

typedef Antichain* AntichainVector;

extern bool containment;


//// Letters of the alphabet 
extern int num_letters;


extern int sizeA;
extern bool* initialA;
extern bool* acceptingA;
extern bool*** transitionsA; // Transitions over all letters 

////the second, which we will 
extern int sizeB;
extern int maxRank;
extern bool* initialB;
extern bool* acceptingB;
extern bool*** transitionsB; // Transitions over all letters 

extern int largestRank;


extern List** postA;
extern List** postB;
extern List** preA;

extern int num_inserts;
extern int num_removed;
extern int num_subsumed;


double get_elapsed_time();

// SF: Constructors and destructors
Set emptySet(); 

// returns true if <s> is empty
bool isEmptySet(Set s);

void clearSet(Set s);

Set initialSet();

Set copySet(Set s);

//Triple* createTriple(State a, Set s, Set o);

Pair* createPair(Set s, Set o);

Pair* copyPair(Pair* t);

void deletePair(Pair* p);
void deleteSet(Set p);

int min(int x, int y);

int ceilOdd(int n);

Pair* intersectPair(Pair *t1, Pair *t2);
Set intersectSet(Set t1, Set t2);

bool LEQpair(Pair* p, Pair* q);
bool EQpair(Pair* p, Pair* q);
bool EQset(Set p, Set q);

void printWord(Word* W);

void printList(List list);

void printAllPost();

void printSet(Set s);

void printSmartSet(Set s);

void printPair(Pair* pair);

void printSmartPair(Pair* pair);

void printAntichain(Antichain* Q);

void printSmartAntichain(Antichain* Q);

Counterexample* generateCounterexample(Antichain* init, Antichain* y);

#endif
